/* wakeup version 0.2 */

/*
 * Copyright (c) 2005-2006 Markus Hennecke
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the Author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR 
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/types.h>

#include <libnet.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

u_int8_t enet_dst[6] = { 0xff, 0xff, 0xff, 0xff, 0xff, 0xff };
u_int8_t enet_src[6] = { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };
u_int8_t mac[6] = { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };

/*
 * Parses a mac address supplied as a zero terminated string via the s
 * argument and writes the address to the memory specified via the mac
 * argument.
 * Returns 0 on failure.
 */
int
parse_mac(const char *s, u_int8_t *mac)
{
	int i, j, pos;
	char c;

	/* Check for the right lengths */
	if (strlen(s) != 17)
		return 0;

	j = 0;
	pos = 0;
	for (i = 0; i < 17; i++) {
		c = toupper(s[i]);
		switch (i) {
		case 2:
		case 5:
		case 8:
		case 11:
		case 14:
			if ((c != ':') && (c != '-'))
				return 0;
			break;
		default:
			if (! isxdigit(c))
				return 0;

			c -= 0x30;
			if (c > 9)
				c -= 7;

			mac[pos] = (j % 2) ? (mac[pos] | c) : (c << 4);

			if (j++ % 2)
				pos++;
		}
	}
			
	return 1;
}

/*
 * Prints out the usage information. 
 * The function will not return but exit with exit code 1.
 */
__dead void
usage(void)
{
	fprintf(stderr, "USAGE: wakeup interface mac\n");
	exit(1);
}

/*
 * Main function
 */
int
main(int argc,  char **argv)
{
	int ret = 0;
	const size_t payload_len = 16 * 6;
	u_int8_t *payload = NULL;

	int i;

	char *device = NULL;
	libnet_t *context;
	u_int32_t source_address, destination_address = 0xffffffff;

	libnet_ptag_t udp_tag, ipv4_tag, ethernet_tag;

	char err_buf[LIBNET_ERRBUF_SIZE];

	/* Parse the command line arguments */
	if (argc != 3)
		usage();
	
	if (! parse_mac(argv[2], mac))
		usage();

	device = argv[1];

	/* Init the libnet */
	context = libnet_init(LIBNET_LINK, device, err_buf);
	if (context == NULL) {
		/* Display the error message */
		fprintf(stderr, "wakeup: Unable to init libnet\n"
				"%s", err_buf);
		usage();
	}

	/* Get the source ipv4 address */
	source_address = libnet_get_ipaddr4(context);
	if (source_address == 0xffffffff) {
		/* Display the error message */
		fprintf(stderr,
			"wakeup: Unable to retrieve ip address from %s\n"
			"\t%s", device, libnet_geterror(context));
		ret = 2;
		goto bailout;
	}

	/* Build the payload */
	payload = malloc(payload_len);
	if (! payload) {
		fprintf(stderr, "wakeup: Unable to allocate memory\n");
		ret = 3;
		goto bailout;
	}

	for (i = 0; i < 16; i++) {
		memcpy(&payload[i * 6], mac, 6);
	}

	/* Create the packet */
	udp_tag = libnet_build_udp(40000, 9, LIBNET_UDP_H + payload_len,
				   0, payload, payload_len, context, 0);
	if (udp_tag == -1) {
		fprintf(stderr, 
			"wakeup: Unable to create udp packet\n"
			"\t%s", libnet_geterror(context));
		ret = 4;
		goto bailout;
	}

	ipv4_tag = libnet_build_ipv4(LIBNET_IPV4_H + LIBNET_UDP_H + payload_len,
				     0x10, 242, 0, 1, IPPROTO_UDP, 0,
				     source_address, destination_address, 
				     NULL, 0, context, 0);
	if (ipv4_tag == -1) {
		fprintf(stderr, 
			"wakeup: Unable to create ip packet\n"
			"\t%s", libnet_geterror(context));
		ret = 5;
		goto bailout;
	}

	ethernet_tag = libnet_autobuild_ethernet(enet_dst, ETHERTYPE_IP,
						 context);
	if (ethernet_tag == -1) {
		fprintf(stderr, 
			"wakeup: Unable to create ip packet\n"
			"\t%s", libnet_geterror(context));
		ret = 6;
		goto bailout;
	}

	/* Write the packet to the wire */
	ret = libnet_write(context);
	if (ret == -1) {
		fprintf(stderr, "wakeup: Error writing wakeup packet\n"
			"\t%s", libnet_geterror(context));
		ret = 7;
	} else 
		ret = 0;

bailout:
	/* Cleanup */
	libnet_destroy(context);

	if (payload)
		free(payload);

	return ret;
}

